"""sync

Implements the sync command of jotter.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
import git

from jotter import util


def run() -> None:
    """Run the sync command."""
    if not util.jotter_project_exists():
        print("Unable to sync, the current folder is not a jotter project.")
        raise SystemExit(1)

    repo = git.Repo(path=".")
    if "origin" not in repo.remotes:
        print("Unable to sync, repo did not have remote. "
              "Have you run jotter set-remote?")
        raise SystemExit(1)

    print("Syncing...")
    try:
        if len(repo.remote().refs) > 0:
            repo.remote().pull("master")
        repo.remote().push("master")
    except git.GitError as err:
        print(f"Unable to sync, {err}")
        raise SystemExit(1)



"""util

Contains various utility functions for jotter.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
from os import path

from jotter import project


def jotter_project_exists() -> bool:
    """Ensure that a jotter project exists in the current path.

    Checks for a git repository (.git folder), and a jotter project file.

    Returns:
        True if a project exists, False otherwise.
    """
    return path.exists(".git") and path.exists(project.PROJECT_FILENAME)


"""note

Functionality for the jotter note command.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
from os import path
import subprocess

import git

from jotter import util, project


def run(note_filename: str) -> None:
    """Run the note command.

    Args:
        project: The jotter project.
        note_filename: The filename of the note.
    """
    if not util.jotter_project_exists():
        print(
            "Unable to create note, the current folder is not a jotter project."
        )
        raise SystemExit(1)
    proj = project.load(project.PROJECT_FILENAME)

    # if note doesn't have file extension, add the one specified in project
    if "." not in note_filename:
        note_filename += f".{proj.note_extension}"

    # if the node doesn't exist, create it first
    editing = True
    if not path.exists(note_filename):
        editing = False
        create_note(note_filename)

    # launch subprocess to edit
    try:
        editor = proj.get_editor()
    except KeyError:
        raise SystemExit(1)
    subprocess.run([editor, note_filename])

    try:
        repo = git.Repo(path=".")
        repo.index.add([note_filename])
        commit_action_word = "Edit" if editing else "Create"
        repo.index.commit(f"{commit_action_word} {note_filename}")
    except git.GitError as err:
        print(f"Unable to create note, {err}")
        raise SystemExit(1)


def create_note(note_filename: str) -> None:
    """Create a note at a given file path.
    
    Exits program if there was some error.
    """
    try:
        open(note_filename, "w").close()
    except OSError as err:
        print(f"Unable to create note, {err}")
        raise SystemExit(1)


"""project

Contains schemas and code for a jotter project file.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
from dataclasses import dataclass, field
import platform
from typing import Mapping

from marshmallow import Schema, fields, post_load, ValidationError, EXCLUDE
import yaml

PROJECT_FILENAME = "jotterproject.yml"
"""The filename of a jotter project."""

DEFAULT_EDITORS = {"Windows": "notepad", "Darwin": "vi", "Linux": "vi"}
"""The default editors for the default platforms (Windows, OSX, Linux)."""


def default_editors() -> Mapping[str, str]:
    """Factory function for the default editor value in a Project."""
    return DEFAULT_EDITORS


class ProjectSchema(Schema):
    """Marshmallow schema for a Project object."""
    editor = fields.Mapping(keys=fields.Str,
                            values=fields.Str,
                            required=True,
                            allow_none=False)
    note_extension = fields.Str(required=True, allow_none=False, default="md")

    @post_load
    def make_project(self, data, **kwargs):
        return Project(**data)


@dataclass
class Project:
    """A jotter project.

    Attributes:
        editor: Mapping between platform.system() and default editor (i.e. 'vi').
        note_extension: File extension of notes. Default is 'md'.
    """
    editor: Mapping[str, str] = field(default_factory=default_editors)
    note_extension: str = "md"

    def get_editor(self) -> str:
        """Gets the editor command for the current platform.

        Raises:
            KeyError: If the platform wasn't listed in the editor dict.
        """
        current_platform = platform.system()
        try:
            return self.editor[current_platform]
        except KeyError:
            print(f"Unable to get editor for platform '{current_platform}'. "
                  "Please add it to the 'editor' dictionary in your project.")
            raise


def _print_validation_err(err: ValidationError, name: str) -> None:
    """Internal function used for printing a validation error in the Schema.
    Args:
        err (ValidationError): The error to log.
        name (str): A human-readable identifier for the Schema data source. 
            Like a filename.
    """
    # build up a string for each error
    log_str = []
    log_str.append(f"Error validating config '{name}':")
    for field_name, err_msgs in err.messages.items():
        log_str.append(f"{field_name}: {err_msgs}")

    # print the joined up string
    print(" ".join(log_str))


def load(project_path: str) -> Project:
    """Load a project from a YAML file.

    Args:
        project_path: The path to the project file.

    Returns:
        The loaded Project, or None if some error occurred.
    """
    try:
        with open(project_path, "r") as project_file:
            raw_project = yaml.safe_load(project_file)
            loaded_project = ProjectSchema(unknown=EXCLUDE).load(raw_project)
            return loaded_project
    except OSError as err:
        # there was some error opening the file
        print(f"Error opening project file: {err}")
        return None
    except yaml.YAMLError as err:
        # the YAML was malformed
        print(f"Error in project file: {err}")
        return None
    except ValidationError as err:
        # the project file was invalid
        _print_validation_err(err, project_path)
        return None


def dump(project: Project, project_path: str) -> None:
    """Dump a project to a YAML file.

    Args:
        project: The Project object to dump.
        project_path: The path to dump the project to.

    Raises:
        OSError: If there was some error writing the file.
    """
    try:
        with open(project_path, "w") as project_file:
            dumped_project = ProjectSchema().dump(project)
            yaml.safe_dump(dumped_project, project_file)
    except OSError as err:
        # there was some error writing the file
        print(f"Error writing project file: {err}")
        raise


"""set_remote

The functionality for the jotter set-remote command.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
import git

from jotter import util


def run(remote_url: str) -> None:
    """Run the set-remote command.
    
    Args:
        remote_url: The remote URL to use.
    """
    if not util.jotter_project_exists():
        print(
            "Unable to set remote, the current folder is not a jotter project."
        )
        raise SystemExit(1)

    repo = git.Repo(path=".")
    print("Setting remote...")
    if "origin" in repo.remotes:
        repo.remote("origin").set_url(remote_url)
    else:
        repo.create_remote("origin", remote_url)


"""init

Contains jotter functionality specific to the init command.

Author
    Figglewatts <me@figglewatts.co.uk>
"""
from os import path
from typing import Optional

import git

from jotter import project, util


def run(remote_url: Optional[str] = None) -> None:
    """Run the init command.

    Will exit the program if the current folder is already a git repo.

    Args:
        remote_url: The optional remote URL to clone from.
    """
    if util.jotter_project_exists():
        print("Unable to init project, this folder is already a jotter "
              "project or git repo.")
        raise SystemExit(1)

    if not remote_url:
        create_new()
    else:
        clone_existing(remote_url)


def create_new() -> None:
    """Create a new jotter project in the current directory."""
    if path.exists(".git"):
        print(
            "Unable to create new project, this folder is already a git repo.")
        raise SystemExit(1)

    print("Creating project...")
    proj = project.Project()
    project.dump(proj, project.PROJECT_FILENAME)

    print("Creating repository...")
    repo = git.Repo.init()

    print("Creating commit...")
    repo.index.add(repo.untracked_files)
    repo.index.commit("Initial commit")
    repo.create_head("master")


def clone_existing(remote_url: str) -> None:
    pass


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jotter']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'gitpython>=3.1.7,<4.0.0',
 'marshmallow>=3.7.1,<4.0.0',
 'pyyaml>=5.3.1,<6.0.0']

setup_kwargs = {
    'name': 'jotter',
    'version': '0.1.0',
    'description': 'Note taking and syncing utility.',
    'long_description': None,
    'author': 'Figglewatts',
    'author_email': 'figglewatts@hotmail.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""cli

The main CLI and entrypoint of jotter.

Author:
    Figglewatts <me@figglewatts.co.uk>
"""
from typing import Optional

import click

import jotter as jotter_pkg
from jotter import init as init_cmd
from jotter import set_remote as set_remote_cmd
from jotter import sync as sync_cmd
from jotter import note as note_cmd

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
"""Context settings for Click."""


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(jotter_pkg.__version__)
def jotter():
    """Note taking and syncing utility."""


@jotter.command()
@click.argument("REMOTE_URL", required=False, type=str)
def init(remote_url: Optional[str] = None) -> None:
    """Initialise a new jotter project in the current directory."""
    init_cmd.run(remote_url)


@jotter.command()
@click.argument("REMOTE_URL", required=True, type=str)
def set_remote(remote_url: str) -> None:
    """Set the remote URL of the jotter project, for syncing."""
    set_remote_cmd.run(remote_url)


@jotter.command()
@click.argument("NOTE_FILE", required=True, type=str)
def note(note_file: str) -> None:
    """Create or edit a jotter note."""
    note_cmd.run(note_file)


@jotter.command()
def sync() -> None:
    """Sync this jotter project with the remote."""
    sync_cmd.run()


def main() -> None:
    jotter()


if __name__ == "__main__":
    main()